/*
   Projet IFT763 : Objets volant avec Verlet / Contraintes

   (C) Eric Beaudry 2004.
   Eric.Beaudry@USherbrooke.ca
   http://siela.dmi.usherb.ca/~eric/

   Universit de Sherbrooke
   Facult des Sciences
   Dpartement d'Informatique

*/
#include <iostream>
#include "physic.h"


Particule::Particule()
: mass(0.0), id(-1), isVirtual(false){
}

Nail::Nail(Particule* p_)
: fixpos(p_->pos), p(p_)
{
}

Spring::Spring(Particule* p1_, Particule* p2_, double k_, double d_)
: p1(p1_), p2(p2_), k(k_), d(d_)
{
   rest = p1->pos.distance(p2->pos);
}

Stick::Stick(Particule* p1_, Particule* p2_)
: p1(p1_), p2(p2_){
   distance = p1->pos.distance(p2->pos);
   totalInvMass = 1.0/p1->mass + 1.0/p2->mass;
}

SurfaceTriangle::SurfaceTriangle(Particule* p1_, Particule* p2_, Particule* p3_)
: p1(p1_), p2(p2_), p3(p3_)
{
}

//--------------

PhysicSimulator::PhysicSimulator(double g, int iterPerSec)
: forceG(g), terrain(NULL), forces(NULL)
{
   deltaT = 1.0 / iterPerSec;
   airDensity = 1.294217666;
   //wind = Vector(10.00, 1.01, 2.0);
}

PhysicSimulator::~PhysicSimulator(){
   delete[]  forces;
}

void PhysicSimulator::addParticule(Particule* particule){
   particules.push_back(particule);
}

void PhysicSimulator::addNail(const Nail& nail){
   nails.push_back(nail);
}

void PhysicSimulator::addSpring(const Spring& r){
   springs.push_back(r);
}

void PhysicSimulator::addSurface(const SurfaceTriangle& surf){
   surfaces.push_back(surf);
}

void PhysicSimulator::setWind(const Vector& wind_){
   wind = wind_;
}

void PhysicSimulator::addWind(const Vector& change){
   wind += change;
}

void PhysicSimulator::setTerrain(Terrain* terrain_){
   terrain = terrain_;
}

void PhysicSimulator::init(){
   forces = new Vector[particules.size()];
   for(unsigned int i=0;i<particules.size();i++){
      particules[i]->id = i;
   }
}

void PhysicSimulator::applyForces(){
   // Apply Gravity
   for(unsigned int i=0;i<particules.size();i++){
       forces[i].z -= particules[i]->mass * forceG;
   }
}


// ---------------------------
PhysicSimulatorEuler::PhysicSimulatorEuler(double g, int iterPerSec)
: PhysicSimulator(g, iterPerSec), velocity(NULL)
{
}

PhysicSimulatorEuler::~PhysicSimulatorEuler(){
   delete[] velocity;
}

void PhysicSimulatorEuler::init(){
   PhysicSimulator::init();
   velocity = new Vector[particules.size()];
}

void PhysicSimulatorEuler::applyForces(){
   PhysicSimulator::applyForces();
   // Springs
   for(unsigned int i=0;i<springs.size();i++){
       Vector L = springs[i].p1->pos - springs[i].p2->pos;
       Vector v = velocity[springs[i].p1->id] - velocity[springs[i].p2->id];
       double l = L.norm();
       double a = springs[i].k * (l - springs[i].rest);
       double b = springs[i].d * (v*L) /l;
       Vector F = L * (a + b) / l;
       forces[springs[i].p1->id] -= F;
       forces[springs[i].p2->id] += F;
   }

   // Surfaces / Wind
   for(unsigned int i=0;i<surfaces.size();i++){
       Vector v1 = surfaces[i].p2->pos - surfaces[i].p1->pos;
       Vector v2 = surfaces[i].p3->pos - surfaces[i].p1->pos;
       Vector force = v1 ^ v2;
       Vector vel = velocity[surfaces[i].p1->id]
                  + velocity[surfaces[i].p2->id]
                  + velocity[surfaces[i].p3->id];
       vel /= 3;
       vel -= wind;
       double WrelDotCAxBA = vel * force;
       double sign = WrelDotCAxBA > 0? -1: 1;
       double force2 = force.norm2();
       if(force2 > 0) {
          double k = sign*airDensity*WrelDotCAxBA*WrelDotCAxBA/force2;
          force *= k;
          //std::cout << "vel=" << vel << "  force=" << force << std::endl;
          forces[surfaces[i].p1->id] += force;
          forces[surfaces[i].p2->id] += force;
          forces[surfaces[i].p3->id] += force;
       }
    }
}


void PhysicSimulatorEuler::iter(){

    applyForces();

    // Velocity
    for(unsigned int i=0;i<particules.size();i++){
       velocity[i] += forces[i] / particules[i]->mass * deltaT;
       forces[i].setZero();
    }

    // Position
    for(unsigned int i=0;i<particules.size();i++){
       particules[i]->pos += velocity[i] * deltaT;
       double minz = terrain ? terrain->getLevel(particules[i]->pos.x, particules[i]->pos.y) : 0;
       if(particules[i]->pos.z <= minz){
          particules[i]->pos.z = minz;
          velocity[i].z = 0;
          velocity[i] *= 0.979; // small friction factor
       }
    }

    // Nails
    for(unsigned int i=0;i<nails.size();i++){
       nails[i].p->pos = nails[i].fixpos;
       velocity[i].setZero();
    }
}

Vector PhysicSimulatorEuler::getVelocity(const Particule* p) const{
   return velocity[p->id];
}

// -------------------

PhysicSimulatorVerlet::PhysicSimulatorVerlet(double g, int iterPerSec)
: PhysicSimulator(g,iterPerSec), oldPos(NULL), nbLinkPart(NULL)
{
}

PhysicSimulatorVerlet::~PhysicSimulatorVerlet(){
   delete[] nbLinkPart;
   delete[] oldPos;
}

int PhysicSimulatorVerlet::addStick(const Stick& s){
   const int i = (int) sticks.size();
   sticks.push_back(s);
   return i;
}

void PhysicSimulatorVerlet::init(){
   PhysicSimulator::init();
   nbLinkPart = new int[particules.size()];
   oldPos = new Point[particules.size()];
   for(unsigned int i=0;i<particules.size();i++){
      oldPos[i] = particules[i]->pos;
      nbLinkPart[i] = 0;
   }
   for(unsigned int i=0;i<sticks.size();i++){
      nbLinkPart[sticks[i].p1->id]++;
      nbLinkPart[sticks[i].p2->id]++;
   }
   for(unsigned int i=0;i<particules.size();i++){
     //std::cout << "Particule[" << i << "] = " << particules[i]->pos.x << " " << particules[i]->pos.y << " " << particules[i]->pos.z << std::endl;
   }
   /*for(unsigned int i=0;i<sticks.size();i++)
     std::cout << "Sticks[" << i << "] = " << sticks[i].p1->id << " - " << sticks[i].p2->id << std::endl;
   for(unsigned int i=0;i<surfaces.size();i++)
     std::cout << "Triangle[" << i << "] = " << surfaces[i].p1->id << " - " << surfaces[i].p2->id << " - " << surfaces[i].p3->id << std::endl;
   for(unsigned int i=0;i<nails.size();i++)
     std::cout << "Nail[" << i << "] : " << nails[i].p->id << std::endl; */
}

void PhysicSimulatorVerlet::applyForces(){
   PhysicSimulator::applyForces();  // this apply gravity

   // Springs
   for(unsigned int i=0;i<springs.size();i++){
       Vector L = springs[i].p1->pos - springs[i].p2->pos;
       Vector v = (springs[i].p1->pos - oldPos[springs[i].p1->id]) / deltaT   // Velocity P1
                - (springs[i].p2->pos - oldPos[springs[i].p2->id]) / deltaT;  // Velocity P2
       double l = L.norm();
       double a = springs[i].k * (l - springs[i].rest);
       double b = springs[i].d * (v*L) /l;
       Vector F = L * (a + b) / l;
       forces[springs[i].p1->id] -= F;
       forces[springs[i].p2->id] += F;
   }

   // Surfaces / Wind
   for(unsigned int i=0;i<surfaces.size();i++){
       Vector v1 = surfaces[i].p2->pos - surfaces[i].p1->pos;
       Vector v2 = surfaces[i].p3->pos - surfaces[i].p1->pos;
       Vector force = v1 ^ v2;
       Vector wvel = (surfaces[i].p1->pos - oldPos[surfaces[i].p1->id]) // Velocity P1
                   + (surfaces[i].p2->pos - oldPos[surfaces[i].p2->id]) // Velocity P2
                   + (surfaces[i].p3->pos - oldPos[surfaces[i].p3->id]); // Velocity P3
       wvel /= (3 * deltaT);  // to get mean
       wvel -= wind;          // remove wind : we get force;
       double WrelDotCAxBA = wvel * force;
       double sign = WrelDotCAxBA > 0? -1: 1;
       double force2 = force.norm2();
       if(force2 > 0) {
          double k = sign*airDensity*WrelDotCAxBA*WrelDotCAxBA/force2;
          force *= k;
          //std::cout << "vel=" << vel << "  force=" << force << std::endl;
          forces[surfaces[i].p1->id] += force;
          forces[surfaces[i].p2->id] += force;
          forces[surfaces[i].p3->id] += force;
       }
    }
}

Vector PhysicSimulatorVerlet::getVelocity(const Particule* p) const{
   return Vector(oldPos[p->id], p->pos) / deltaT;
}

void PhysicSimulatorVerlet::iter(){

    applyForces();

    double  dt2 = deltaT * deltaT;

    // Verlet Integration
    for(unsigned int i=0;i<particules.size();i++){
        oldPos[i].x = (particules[i]->pos.x*2.0) - oldPos[i].x;
        oldPos[i].y = (particules[i]->pos.y*2.0) - oldPos[i].y;
        oldPos[i].z = (particules[i]->pos.z*2.0) - oldPos[i].z;
        oldPos[i] += forces[i] * (dt2 / particules[i]->mass);
        forces[i].setZero();
    }

    // Satisfy constraints
    satisfyConstraints();

    // Make effective new positions
    for(unsigned int i=0;i<particules.size();i++){
        Point temp = oldPos[i];
        if(!particules[i]->isVirtual){
           double minz = terrain ? terrain->getLevel(temp.x, temp.y) : 0;
           if(temp.z <= minz){
               temp.z = minz;
               particules[i]->pos = temp;
           }
        }
        oldPos[i] = particules[i]->pos;
        particules[i]->pos = temp;
    }
}

void PhysicSimulatorVerlet::move(const Vector& d) {
  for(unsigned int i=0;i<particules.size();i++){
     particules[i]->pos += d;
     oldPos[i] = particules[i]->pos;
  }
}

void PhysicSimulatorVerlet::satisfyConstraints(){
   for(int i=0;i<55;i++){
       satisfySticks();
       meanForces();
       satisfyNails();
   }
}

void PhysicSimulatorVerlet::satisfySticks(){
   unsigned int n = (unsigned int) sticks.size();
   for(unsigned int i=0;i<n;i++){
       Particule& p1 = *sticks[i].p1;
       Particule& p2 = *sticks[i].p2;
       Vector v = oldPos[p1.id] - oldPos[p2.id];
       double l = v.norm(); //  amliorer pour viter la sqrt (voir approximation)
       double d = (l-sticks[i].distance) / (l * sticks[i].totalInvMass);
       forces[p1.id] -= v * (d / p1.mass);
       forces[p2.id] += v * (d / p2.mass);
       //double d = (l-sticks[i].distance);
       //double l2 = sticks[i].distance*sticks[i].distance;
       //double d= (-l2+v.norm2())/(l2+v.norm2());
       //forces[p1.id] -= v * (d * 0.5);
       //forces[p2.id] += v * (d * 0.5);
   }
}

void PhysicSimulatorVerlet::meanForces(){
   unsigned int n = (unsigned int) particules.size();
   for(unsigned int i=0;i<n;i++){
      if(nbLinkPart[i])
          oldPos[i] += forces[i] / nbLinkPart[i];
      forces[i].setZero();
   }
}

void PhysicSimulatorVerlet::satisfyNails(){
   for(unsigned int i=0;i<nails.size();i++)
      oldPos[nails[i].p->id] = nails[i].fixpos;
}

#include <GL/glut.h>
void PhysicSimulatorVerlet::renderGL(){
  glColor3d(1, 0, 0);
  //glEnable(GL_POINT_SMOOTH);
  glEnable(GL_LINE_SMOOTH);
  glPointSize(1);
  glLineWidth(3);
  glBegin(GL_POINTS);
  for(unsigned int i=0;i<particules.size();i++){
     Point& p = particules[i]->pos;
     glVertex3d(p.x, p.y, p.z);
  }
  glEnd();

  glBegin(GL_LINES);
  for(unsigned int i=0;i<sticks.size();i++){
     Stick& s = sticks[i];
     Point& p1 = s.p1->pos;
     Point& p2 = s.p2->pos;
     glVertex3d(p1.x, p1.y, p1.z);
     glVertex3d(p2.x, p2.y, p2.z);
  }
  glEnd();

  // Surfaces...
  glBegin(GL_TRIANGLES);
  glColor3d(0, 0.7, 0.5);
  for(unsigned int i=0;i<surfaces.size();i++){
     SurfaceTriangle& st = surfaces[i];
     Point& p1 = st.p1->pos;
     Point& p2 = st.p2->pos;
     Point& p3 = st.p3->pos;
     Vector vn = (p2-p1) ^ (p3-p1);
     vn.normalize();
     glNormal3d(vn.x, vn.y, vn.z);
     glVertex3d(p1.x, p1.y, p1.z);
     glVertex3d(p2.x, p2.y, p2.z);
     glVertex3d(p3.x, p3.y, p3.z);
  }
  glEnd();
}
