/*
   Projet IFT763 : Objets volant avec Verlet / Contraintes

   (C) Eric Beaudry 2004.
   Eric.Beaudry@USherbrooke.ca
   http://siela.dmi.usherb.ca/~eric/

   Universit de Sherbrooke
   Facult des Sciences
   Dpartement d'Informatique

*/

#if !defined(__PHYSIC_H__)
#define __PHYSIC_H__

#include <vector>
#include "geometrie.h"
#include "terrain.h"


class Particule {
   public:
      Particule();

      Point        pos;
      double       mass;
      int          id;

      bool         isVirtual;
};

class SurfaceTriangle{
   public:
     SurfaceTriangle(Particule* p1, Particule* p2, Particule* p3);

   protected:
     Particule*    p1;
     Particule*    p2;
     Particule*    p3;

   friend class PhysicSimulatorEuler;
   friend class PhysicSimulatorVerlet;
};

class Nail {
   public:
      Nail(Particule* p);

   protected:
      Point        fixpos;
      Particule*   p;

   friend class PhysicSimulatorEuler;
   friend class PhysicSimulatorVerlet;
};

class Spring{
   public:
      Spring(Particule* p1, Particule* p2, double k=100, double d=1);

   protected:
      Particule*        p1;
      Particule*        p2;
      double            rest, k, d;

   friend class PhysicSimulatorEuler;
   friend class PhysicSimulatorVerlet;
};

class Stick{
   public:
      Stick(Particule* p1, Particule* p2);
      //Stick();

      void              setDistance(double d) { distance = d;}
      double            getDistance() const { return distance;}

   protected:
      Particule*        p1;
      Particule*        p2;
      double            distance;
      double            totalInvMass;

   friend class PhysicSimulatorVerlet;
};

class PhysicSimulator {
   public:
      PhysicSimulator(double g, int iterPerSec);
      virtual ~PhysicSimulator();

      void          addParticule(Particule* particule);
      void          addNail(const Nail&);
      void          addSpring(const Spring&);
      void          addSurface(const SurfaceTriangle&);

      void          setWind(const Vector& wind);
      Vector&       getWind() { return wind;}
      void          addWind(const Vector& change);

      void          setTerrain(Terrain* terrain);

      virtual void  applyForces();
      virtual void  init();
      virtual void  iter() = 0;

      virtual Vector getVelocity(const Particule* particule) const = 0; // Get the current velocity

   protected:
      double        deltaT;
      double        forceG;
      Vector        wind;
      double        airDensity;

      Terrain*      terrain;  //for ground level

      std::vector<Particule*>      particules;
      std::vector<Nail>            nails;
      std::vector<Spring>          springs;
      std::vector<SurfaceTriangle> surfaces;

      Vector*       forces;
};


class PhysicSimulatorEuler : public PhysicSimulator {
   public:
      PhysicSimulatorEuler(double g=9.80665, int iterPerSec=100);
      virtual ~PhysicSimulatorEuler();

      virtual void   applyForces(); // Springs + Wind
      virtual void   init();
      virtual void   iter();

      virtual Vector getVelocity(const Particule* particule) const;

   protected:
      Vector*        velocity;
};

class PhysicSimulatorVerlet : public PhysicSimulator{
   public:
      PhysicSimulatorVerlet(double g=9.80665, int iterPerSec=100);
      virtual ~PhysicSimulatorVerlet();

      int            addStick(const Stick&);

      Stick&         getStick(int index) { return sticks[index];}

      virtual void   applyForces(); // Springs + Wind
      virtual void   init();
      virtual void   iter();

      virtual Vector getVelocity(const Particule* particule) const;

      void  move(const Vector& d);

      void           renderGL();

   protected:
      void          satisfyConstraints();
      void          satisfySticks();
      void          meanForces();
      void          satisfyNails();

      std::vector<Stick>           sticks;
      Point*                       oldPos;
      int*                         nbLinkPart;

};

#endif
